package de.ims.cl.dv.parsers;

import java.io.*;
import java.util.*;
import java.util.regex.*;

import de.ims.cl.dv.datastructures.*;
import de.ims.cl.dv.util.*;

public class MiniparCorpus implements Corpus {
    
    String fixlemma(String lemma){
        if (lemma.equals("")) return "_";
        return lemma;
    }
    
    
    BufferedReader corpus;
    Logfile log;
    long charcounter = 0; // count how much has been read from the corpus

    Graph nextParse; // lookahead
    
    Comparator<Edge> miniparComparator = new MiniparEdgeComparator();
    
    Pattern punctuation = Pattern.compile("^[^\\w]+$");
    
    enum ParseTypes {TRIPLES, FULL, BARONI, UNDETERMINED};
    
    ParseTypes parseType;
    
    private void increaseCounter(int inc) throws IOException { // increase charcounter and output
    long distance = 1000000;
    if ((charcounter/distance -
         (charcounter +=inc)/distance) < 0) { // printout every 10 mill lines
        log.println("Reading corpus byte "+(new Long(charcounter)).toString());
    } 
    }
    
    /**
       Returns the number of bytes read from the current corpus
    **/
    
    public long charcounter() {
    return charcounter;
    }
  
    
    // read a sentence in MINIPAR "full" output
    //
    // This is the format where missing parent nodes are denoted as *,
    // nodes without terminals as (), and where terminal nodes with function words 
    // or punctuation are occasionally missing.

    // Since edges may refer to nodes have not been seen yet, we at first model all
    // edges as PreliminaryEdges; these are turned into proper Edges once the whole 
    // parse graph has been parsed.


    private Graph fetchNextPGFull() throws IOException {
    Graph pg = new Graph(); 
    List<PreliminaryEdge> edges = new LinkedList<PreliminaryEdge>(); // store preliminary edges
    Map<Integer,Integer> antecedents = new HashMap<Integer,Integer>();
    String line,lemma,lemma_pos,rel;
    Integer lemmaIndex, govIndex;
    boolean error = false;
    String[] field3;
    while (true) {
        line = corpus.readLine(); // begin reading when you've seen (
        if (line == null) {// EOF
                    // have we created a non-empty graph up to now? then return, else return null to signal EOF
            if (pg.isEmpty()) {
                return null; 
            } else {
                return pg;
            }
        }
        // continue here if not EOF    
        increaseCounter(line.length()); // increase counter 
        if (line.startsWith("> ")) {
            line = line.substring(2); // delete trailing >
        }
        if (line.equals("(")) {
            break;
        }
    }
    while (true) {
        line = corpus.readLine();
        if (line == null) {return null;}
        increaseCounter(line.length()); // increase counter 
        if (line.equals(")")) {   // read until you've seen )
            if (error) {
            return new Graph();
            } else {
                // do postprocessing on graph
                // enter edges from preliminary edges List and resolve anaphors 
                for (PreliminaryEdge pe: edges) {               
                    pg.addEdge(pe.constructProperEdge(pg), miniparComparator);
                }
    
                for (Integer antFrom: antecedents.keySet()) {
                    Integer to =  antecedents.get(antFrom);
                    while (antecedents.containsKey(to)) {
                        to = antecedents.get(to);
                    }
                    pg.replaceBy(antFrom,to);
                }
                return pg;
            }
        }

        String[] tokens = line.split("\t");

        lemmaIndex = new Integer(tokens[0].replaceAll("E","100"));
        tokens[1] = tokens[1].substring(1,tokens[1].length()); // remove (

        switch (tokens.length) {
            case 5: // Grammatical node with unlabelled edge
                field3 = tokens[2].split(" ");
                if (field3.length > 2) { 
                  lemma_pos = field3[field3.length-1];
                } else { 
                  lemma_pos = field3[1];
                }
                pg.addNode(new Node("gram", lemma_pos, lemmaIndex, new MiniparEncoding()));
                if (!(tokens[3].equals("*"))) {
                    govIndex = new Integer(tokens[3].replaceAll("E","100"));
                    edges.add(new PreliminaryEdge(lemmaIndex,govIndex,"nolabel"));
                }
                break;
            case 6: // Lexical node with or without lexical governor
                govIndex = new Integer(tokens[3].replaceAll("E","100"));
                field3 = tokens[2].split(" ");
                lemma_pos = field3[field3.length-1];
                if (field3[0].equals("~") || field3[0].equals("fin")) {
                    if (tokens[1].equals("()")) {             
                        lemma = "gram";
                    } else {
                        lemma = tokens[1];
                    }
                } else {
                    int endPos = tokens[2].lastIndexOf(" ");
                    if (endPos < 0) {
                      log.println("Error in token 2 of "+line);
                    }
                    lemma = tokens[2].substring(0,endPos); // everything but POS
                }
                rel = tokens[4];
                
                lemma = fixlemma(lemma);
                pg.addNode(new Node(lemma, lemma_pos, lemmaIndex, new MiniparEncoding()));
                edges.add(new PreliminaryEdge(lemmaIndex, govIndex, rel));
                break;
            case 7: // nodes with antecedents
                govIndex = new Integer(tokens[3].replaceAll("E","100"));
                field3 = tokens[2].split(" ");
                lemma_pos = field3[field3.length-1];
                if (field3[0].equals("~") || field3[0].equals("fin")) {
                    if (tokens[1].equals("()")) {               
                        lemma = "gram";
                    } else {
                        lemma = tokens[1];}
                } else {
                    int endPos = tokens[2].lastIndexOf(" ");
                    if (endPos < 0) {
                        log.println("Error in token 2 of "+line);
                    }
                    lemma = tokens[2].substring(0,endPos); // everything but POS
                }
                rel = tokens[4];
                lemma = fixlemma(lemma);
                pg.addNode(new Node(lemma, lemma_pos, lemmaIndex, new MiniparEncoding()));
                
                // DEBUG
                // System.out.println("rel = '" + rel + "'");
                
                edges.add(new PreliminaryEdge(lemmaIndex, govIndex, rel));
                if (tokens[6].startsWith("(antecedent ")) {
                    String antecedentIdx = tokens[6].substring(12,tokens[6].length()-2).replaceAll("E","100");
                    antecedents.put(lemmaIndex, new Integer(antecedentIdx)); // remove ))
                } else {
                    log.println("Error: could not parse antecedent!");
                    error = true;
                }
                break;
            default: 
                log.println("Cannot parse line "+line+". Skipping sentence.");
                // now turn preliminary edges into proper edges
                return new Graph();
            }
        }
    }

    // read a sentence in MINIPAR "full full" format
    //
    // This is an extended version of the MINIPAR full format.
    // It uses the literals <emptyToken>, <emptyParent>, and <emptyLemma> instead of the
    // character-based encoding of the "full" format. It also represents all input tokens.
    //
    // Since edges may refer to nodes have not been seen yet, we at first model all
    // edges as PreliminaryEdges; these are turned into proper Edges once the whole 
    // parse graph has been parsed.



    private Graph fetchNextPGBaroni() throws IOException {
    Graph pg = new Graph();
    List<PreliminaryEdge> edges = new LinkedList<PreliminaryEdge>(); // store preliminary edges
    Map<Integer,Integer> antecedents = new HashMap<Integer,Integer>();
    String line,lemma,lemma_pos,rel;
    Integer lemmaIndex, govIndex;
    boolean error = false;
    String[] field3;
        while (true) {
        line = corpus.readLine(); // begin reading when you've seen (
        if (line == null) {return null;} // EOF
        increaseCounter(line.length()); // increase counter 
        if (line.startsWith("> ")) {
            line = line.substring(2); // delete trailing >
        }
        if (line.equals("(")) {
            break; }
        }
        
        while (true) {
        line = corpus.readLine();
        if (line == null) {// EOF
                    // have we created a non-empty graph up to now? then return, else return null to signal EOF
            if (pg.isEmpty()) {
            return null; 
            } else {
            return pg;
            }
        }
        // continue here if not EOF
        increaseCounter(line.length()); // increase counter 
        if (line.equals(")")) {   // read until you've seen )
            if (error) {
            return new Graph();
            } else {

            // do postprocessing on graph
            // enter edges from preliminary edges List and resolve anaphors 

            for (PreliminaryEdge pe: edges) {       
                pg.addEdge(pe.constructProperEdge(pg),miniparComparator);
            }
            
            for (Integer antFrom: antecedents.keySet()) {           
                Integer to = antecedents.get(antFrom);
                while (antecedents.containsKey(to)) {
                to = antecedents.get(to); }
                pg.replaceBy(to,antFrom);
            }
            return pg;
            }
        }
        
        String[] tokens = line.split("\t");

        lemmaIndex = new Integer(tokens[0].replaceAll("E","100"));
        tokens[1] = tokens[1].substring(1,tokens[1].length()); // remove ( in front of 2nd token
        
        switch (tokens.length) {
        case 5: // Grammatical node with unlabelled edge
            field3 = tokens[2].split(" ");
            lemma = field3[0];
            lemma_pos = field3[1];
            if ((lemma.equals("fin") || lemma.equals("<emptyLemma>"))) {lemma = "gram"; };
            // System.err.println("Adding node: "+lemma+" "+lemma_pos+" "+lemmaIndex);
            lemma = fixlemma(lemma);
            pg.addNode(new Node(lemma, lemma_pos, lemmaIndex, new MiniparEncoding()));
            
            if (!(tokens[3].equals("<emptyParent>"))) {
            govIndex = new Integer(tokens[3].replaceAll("E","100"));
            edges.add(new PreliminaryEdge(lemmaIndex,govIndex,"nolabel"));
            }
            break;
        case 6: // Lexical node with or without lexical governor
            govIndex = new Integer(tokens[3].replaceAll("E","100"));
            field3 = tokens[2].split(" ");
            lemma_pos = field3[field3.length-1];
            if (field3[0].equals("~") || field3[0].equals("fin")) {
            if (tokens[1].equals("<emptyToken>")) {             
                lemma = "gram"; }
            else {
                lemma = tokens[1];} }
            else {
            int endPos = tokens[2].lastIndexOf(" ");
            if (endPos < 0) {
                log.println("Error in token 2 of "+line);
            }
            lemma = tokens[2].substring(0,endPos); // everything but POS
            }
            if ((lemma.equals("<emptyLemma>"))) { System.err.println("Skipping node "+lemma); continue; }           
            rel = tokens[4];
            // System.err.println("Adding node: "+lemma+" "+lemma_pos+" "+lemmaIndex);
            lemma = fixlemma(lemma);
            pg.addNode(new Node(lemma, lemma_pos, lemmaIndex, new MiniparEncoding()));
            // System.err.println("Adding edge: "+lemmaIndex+" "+govIndex+" "+rel);
            
            // DEBUG
            // System.out.println("rel = '" + rel + "'");
            
            edges.add(new PreliminaryEdge(lemmaIndex, govIndex, rel));
            break;
        case 7: // nodes with antecedents
            govIndex = new Integer(tokens[3].replaceAll("E","100"));
            field3 = tokens[2].split(" ");
            lemma_pos = field3[field3.length-1];
            if (field3[0].equals("~") || field3[0].equals("fin")) {
            if (tokens[1].equals("<emptyToken>")) {             
                lemma = "gram"; }
            else {
                lemma = tokens[1];} }
            else {
            int endPos = tokens[2].lastIndexOf(" ");
            if (endPos < 0) {
                log.println("Error in token 2 of "+line);
            }
            lemma = tokens[2].substring(0,endPos); // everything but POS
            }
            rel = tokens[4];
            if ((lemma.equals("<emptyLemma>"))) { System.err.println("Skipping node "+lemma); continue; }
            // System.err.println("Adding node: "+lemma+" "+lemma_pos+" "+lemmaIndex);
            lemma = fixlemma(lemma);
            pg.addNode(new Node(lemma, lemma_pos, lemmaIndex, new MiniparEncoding()));
            // System.err.println("Adding edge: "+lemmaIndex+" "+govIndex+" "+rel);
            
            // DEBUG
            // System.out.println("rel = '" + rel + "'");
            
            edges.add(new PreliminaryEdge(lemmaIndex, govIndex, rel));
            if (tokens[6].startsWith("(antecedent ")) {
            String antecedentIdx = tokens[6].substring(12,tokens[6].length()-2).replaceAll("E","100");
            antecedents.put(lemmaIndex, new Integer(antecedentIdx)); // remove ))
            } else {
            log.println("Error: could not parse antecedent!");
            error = true;
            }
            break;
        default: 
            log.println("Cannot parse line "+line+". Skipping sentence.");
            // now turn preliminary edges into proper edges
            return new Graph();
        }
        }
    }

// read a sentence in Minipar "dependency triples" output
    
    private Graph fetchNextPGTriples() throws IOException {
    Graph pg = new Graph(); 
    String line,lemma,gov,triple,lemma_pos,rel,gov_pos;
    int counter = 0;
    Map<String,Integer> indices = new HashMap<String,Integer>();
    boolean error = false;
    Integer lindex,gindex;
        while (true) {
        line = corpus.readLine(); // read next line -- in Java, without termination characters
        if (line == null) { // EOF
                            // have we created a non-empty graph up to now? then return, else return null to signal EOF
            if (pg.isEmpty()) {
            return null; 
            } else {
            return pg;
            }
        }
        
        increaseCounter(line.length()); // increase counter 
        
        if (line.startsWith("> ")) {
            line = line.substring(2); // delete trailing >
        }
        if (line.equals("")) {
            if (error) {            
            return new Graph(); }
            else {
            //          pg.lexicaliseEdges();
            return pg; } }
        else {
            String[] tokens = line.split("\t");
            if (tokens.length != 3) {
            log.println("Warning: line "+line+" does not contain three tokens. Skipping sentence.");
            error = true;
            continue;
            }
            lemma = tokens[0]; 
            triple = tokens[1];
            gov = tokens[2];
            if (lemma.equals("~") || gov.equals("~")) {
            continue;
            }
            String[] tripletokens = triple.split(":");
            if (tripletokens.length != 3) {
            log.println("Warning: line "+line+" does not contain three tokens. Skipping sentence.");
            error = true;
            continue;
            }
            lemma_pos = tripletokens[0];
            rel = tripletokens[1];
            gov_pos = tripletokens[2];

            if (indices.containsKey(gov)) {
            gindex = indices.get(gov);} 
            else {
            gindex = new Integer(counter++);
            indices.put(gov, gindex); }         
            Node govNode = new Node(gov,gov_pos,gindex, new MiniparEncoding());
            
            if (indices.containsKey(lemma)) {
            lindex = indices.get(lemma);} 
            else {
            lindex = new Integer(counter++);
            indices.put(lemma, lindex); }           
            Node lemmaNode = new Node(lemma,lemma_pos, lindex, new MiniparEncoding());
            pg.addNode(govNode);
            pg.addNode(lemmaNode);
            pg.addEdge(new Edge(lemmaNode, govNode,rel),miniparComparator);
        }
        }
    }

    /**
       This method detects the type of MINIPAR output in this.corpus:
       - (a) triple format 
       - (b) full format with empty nodes denoted as "()" and some function words removed
       - (c) "Baroni" format with empty nodes denoted as "<emptyToken>" and all input words present
    **/

    private ParseTypes determineCorpusType() throws IOException {
      this.corpus.mark(800);
      if (!(this.corpus.markSupported())) {  // if stream does not support marking, throw IOException
      throw new IOException("Marking not supported for the current stream -- could not determine corpus type. Exiting.");
      }
      ParseTypes thisParseType = ParseTypes.UNDETERMINED;
      Pattern full = Pattern.compile("\t.+\t.+\t");
      Pattern baroni = Pattern.compile("\t.+\t.+\t<emptyParent>");
      Pattern triple = Pattern.compile("\t.+:.+:.+\t");
      while (thisParseType.equals(ParseTypes.UNDETERMINED)) {
      String nextLine = corpus.readLine();
      if (baroni.matcher(nextLine).find()) {
          thisParseType = ParseTypes.BARONI;
          this.log.println("Detected full extended parse format.");}
      else if (full.matcher(nextLine).find()) {
          thisParseType = ParseTypes.FULL;
          this.log.println("Detected full parse format.");}
      else if (triple.matcher(nextLine).find()) {
          thisParseType = ParseTypes.TRIPLES;
          this.log.println("Detected triples format.");}
      }
      this.corpus.reset();
      return thisParseType; 
  }
    
    /** Read corpus from STDIN and write error output to <i>l</i>.  **/
    
    public MiniparCorpus(Logfile l) throws IOException{
    this.log = l;
    this.corpus = new BufferedReader(new InputStreamReader(System.in));
    this.parseType = determineCorpusType();
    nextParse = null;
    next();
    }

    /** Read corpus from <i>filename</i> and write error output to <i>l</i>. 
    The constructor auto-detects the MINIPAR triple format and full format.
     **/
    public MiniparCorpus(String filename, Logfile l) throws IOException {
        this.corpus = new BufferedReader(new FileReader(filename));
        this.log = l;   
        this.parseType = determineCorpusType();
        nextParse = null;
        next();
    }
    
    public static void main (String [] args)
    {
        // DEBUG
        // System.out.print("I AM MINIPAR!");
    }

	public boolean hasNext() {
		   return (nextParse != null);
	}

	public Graph next() {
		 Graph currentParse = nextParse;
	        try {                switch (parseType) {
	                case BARONI:  nextParse = fetchNextPGBaroni(); break;
	                case FULL:    nextParse = fetchNextPGFull(); break;
	                case TRIPLES: nextParse = fetchNextPGTriples(); break;
	                default: throw new RuntimeException("Unknown corpus type: "+parseType);
	                }
	        } catch (IOException e) {
	                throw new RuntimeException("IOException while reading corpus.");
	        }
	        return currentParse;

	}

	public void remove() {
		throw new UnsupportedOperationException();		
	}
}